/**
 * PilrcEdit - a graphical editor for PilRC files.
 * Copyright (C) 2000 Rob Nielsen
 * rnielsen@wn.com.au
 * http://www.wn.com.au/rnielsen/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 */
package pilrcedit;

import pilrcedit.control.*;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Rectangle;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

public class Form extends FormObject
{
  public static final int SAVEBEHIND=1;
  public static final int NOSAVEBEHIND=2;
  public static final int DEFAULTBTNID=3;
  public static final int MODAL=4;
  public static final int FRAME=5;
  public static final int NOFRAME=6;
  public static final int IFDEF=7;
  public static final int IFNDEF=8;
  public static final int ENDIF=9;
  public static final int POPUPLIST=10;

  public static final int BEGIN=300;
  public static final int END=301;
  public static final int TITLE=302;
  public static final int HELPID=303;
  public static final int MENUID=304;

  public static String DEFAULT_TITLE;

  boolean frame=true;
  boolean modal=false;
  boolean saveBehind=true;
  boolean usable=true;
  String title;
  String help;
  String menu;
  String defaultButton;
  static Font boldFont=Font.getFont(1);
  static Font symbolFont=Font.getFont(3);
  Project project;
  FormObject firstObject=null;
  FormObject lastObject=null;
  Rectangle dirty=new Rectangle(0,0,-1,-1);
  boolean allDirty=false;
  StringBuffer unsupported=null;
  Editor editor;
  Vector listeners;
  Vector triggers=null;
  String ifdef;
  String ifndef;
  Color blue=new Color(51,0,153);

  public Form()
  {
	}

  public Form(Project project)
  {
		ifdef=project.getIfDef();
    ifndef=project.getIfNDef();
    form=this;
		this.project=project;
		xExp.setValue(0);
		yExp.setValue(0);
		widthExp.setValue(160);
		heightExp.setValue(160);
		frame=false;
		title=DEFAULT_TITLE;
		evaluatePosition();
	}

  public Form(Project project,StreamTokenizer st) throws IOException
  {
    ifdef=project.getIfDef();
    ifndef=project.getIfNDef();
		form=this;
		this.project=project;
		readID(st);
		readLocation(st);
		evaluatePosition();
		parseOptions(st);
  }

  public void setEditor(Editor editor)
  {
		this.editor=editor;
	}

  public void requestFocus(FormObject object)
  {
		if (editor!=null)
		  editor.setSelected(object);
	}

  public void repaint()
  {
    if (editor!=null)
      editor.screen.repaint();
  }

  public void setDimensions(String x,String y,String width,String height)
  {
		setAllDirty();
		super.setDimensions(x,y,width,height);
	}

  public Rectangle getDirtyRect()
  {
		if (allDirty)
		  return Screen.fullScreen;
		else
		if (dirty.width==-1)
		  return null;
		dirty.x+=x;
		dirty.y+=y;
		return dirty;
	}

	public void setDirty(Rectangle rect)
	{
		if (rect==null)
		{
			dirty.width=-1;
			allDirty=false;
		}
		else
		if (!allDirty)
			if (dirty.width==-1)
				dirty.setBounds(rect);
			else
				dirty.add(rect);
	}

  public void setDirty(int x,int y,int width,int height)
  {
		if (!allDirty)
			if (dirty.width==-1)
				dirty.setBounds(x,y,width,height);
			else
			{
				dirty.add(x,y);
				dirty.add(x+width,y+height);
			}
	}

  public void setAllDirty()
  {
		allDirty=true;
	}

  public void evaluatePosition()
  {
		super.evaluatePosition();
		FormObject obj=firstObject;
		while(obj!=null)
		{
			obj.evaluatePosition();
			obj=obj.next();
			while (obj!=null&&obj.isDependent)
  			obj=obj.next();
		}
	}

  /**
   * Gets the first form object in this form.  FormObjects form a linked list
   * so the rest of the objects can be obtained by calling next().
   */
  public FormObject getFirstObject()
  {
    return firstObject;
  }

  public FormObject getLastObject()
  {
    return lastObject;
  }

  protected void setFirst(FormObject obj)
  {
		firstObject=obj;
		if (obj==null)
		  lastObject=null;
		else
  		obj.setPrev(null);
	}

  protected void setLast(FormObject obj)
  {
		lastObject=obj;
		if (obj==null)
		  firstObject=null;
		else
  		obj.setNext(null);
	}

  /**
   * Inserts a given object after the given one
   * @param obj the object to insert
   * @param prev the preceeding object
   */
  public void insert(FormObject obj,FormObject prev)
  {
    //System.out.println("insert("+obj+","+prev+")");
		FormObject last=obj;

		// if we are moving from one place to another in the same form
		// remove the chain from the previous position
		if (obj.form==this)
		{
			while(last!=null&&last.hasDependent())
				last=last.next();

      if (firstObject==obj&&lastObject==last)
        setFirst(null);
      else
      if (firstObject==obj)
        setFirst(last.next());
      else
      if (lastObject==last)
        setLast(obj.prev());
      else
        obj.prev().setNext(last.next());
		}
		else
  		obj.form=this;

    if (prev==null)
    {
			if (firstObject==null)
			  setLast(last);
			else
			  firstObject.setPrev(obj);
      setFirst(obj);
		}
		else
		{
			if (prev==lastObject)
			  setLast(last);
			last.setNext(prev.next());
			obj.setPrev(prev);
		}
    notifyListeners();
 	}

  public void printObjects()
  {
		FormObject obj=firstObject;
		while(obj!=null)
		{
			System.out.print(obj.label+(obj==firstObject?"(F)":obj==lastObject?"(L)":"")+",");
			FormObject prev=obj;
			obj=obj.next();
			if (obj!=null&&obj.prev()!=prev)
			  System.out.print("!!");
		}
		System.out.println();
	}

  public void add(FormObject obj)
  {
    //System.out.println("add("+obj+")");
		obj.form=this;
		if (firstObject==null)
		{
			setFirst(obj);
			setLast(obj);
		}
		else
		{
		  obj.setPrev(lastObject);
		  setLast(obj);
		}
    notifyListeners();
	}

  public void remove(FormObject obj)
  {
    //System.out.println("remove("+obj+")");
		obj.form=null;
		if (obj==firstObject)
		{
		  setFirst(obj.next());
		  firstObject.setPrev(null);
		}
		else
		  obj.prev().setNext(obj.next());

		if (obj==lastObject)
		  setLast(obj.prev());
		else
		{
			obj.next().setPrev(obj.prev());
			// if we aren't dependent on our previous and our next is on us, make our next
			// independent
			if (!obj.isDependent&&obj.hasDependent())
			  obj.next().disconnect(false);
			else
			if (obj.hasDependent())
  		  obj.next().evaluatePosition();
		}
		setDirty(obj.getBounds());
    notifyListeners();
	}

  public String toString()
  {
		if (title!=null&&!id.isAbsolute())
	    return id+" - \""+title+"\"";
	  if (title!=null)
	    return "\""+title+"\"";
	  return id.toString();
	}

  public FormObject getObjectAt(int px,int py)
  {
		FormObject obj=lastObject;
		while (obj!=null)
		{
			if (obj.contains(px,py)&&obj.isVisible())
			  return obj;
			obj=obj.prev();
		}
		return null;
	}

  /**
   * Called when a control is moved within a form.  Finds a control that
   * the moving one might latch onto as a dependent.
   * @param move the moving control
   */
  public FormObject getPossibleDependent(FormObject move)
  {
		FormObject obj=firstObject;
		while (obj!=null)
		{
			if ((obj!=move&&!obj.hasDependent()&&!(obj.isDependent&&obj.prev()==move))&&
					(Math.abs(obj.x-move.x)<=joinAmount||
					 Math.abs(move.x-(obj.x+obj.width))<=joinAmount||
					 Math.abs(move.x+move.width-obj.x)<=joinAmount||
					 Math.abs(move.x+move.width-(obj.x+obj.width))<=joinAmount) &&

					(Math.abs(obj.y-move.y)<=joinAmount||
					 Math.abs(move.y-(obj.y+obj.height))<=joinAmount||
					 Math.abs(move.y+move.height-obj.y)<=joinAmount||
					 Math.abs(move.y+move.height-(obj.y+obj.height))<=joinAmount))
      {
        // make sure it's not in our tree of dependents
        FormObject check=move;
        while(check.hasDependent())
        {
          if (check==obj)
            continue;
          check=check.next();
        }
        if (check!=obj)
  				return obj;
      }
		  obj=obj.next();
		}
		return null;
	}

  public void paint(Graphics g)
  {
		Rectangle r=g.getClipBounds();
		g.translate(x,y);
		r.x-=x;
		r.y-=y;
		FormObject poppedlist=null;
		FormObject obj=firstObject;
    Color bgColor=new Color(Preferences.getIntPref(Bitmap.PREF_BGCOLOR));
		while (obj!=null)
		{
		  if (obj.getBounds().intersects(r))
		  {
				if (obj.isVisible())
				{
					if (obj instanceof List&&((List)obj).isPoppedUp())
					  poppedlist=obj;
					else
					{
            g.setColor(bgColor);
						obj.clear(g);
            g.setColor(obj.usable?Color.black:Color.lightGray);
						obj.paint(g);
					}
				}
			}
			obj=obj.next();
		}
		g.translate(-x,-y);
		r.x+=x;
		r.y+=y;

    boolean color=(Preferences.getIntPref(Bitmap.PREF_COLOR_DEPTH,0)==Bitmap.COLOR);
    g.setColor(color?blue:Color.black);

    String drawTitle=(title!=null?TextTools.unescape(title):null);

		if (frame)
		{
			g.fillRect(x-2,y,2,height);
			g.fillRect(x+width,y,2,height);
			g.drawLine(x,y+height-1,x-1,y+height);
			g.drawLine(x+width-1,y+height-1,x+width,y+height);
      g.fillRect(x,y+height,width,2);
			if (r.y<y+15)
			{
				g.fillRect(x,y-2,width,2);
				g.drawLine(x,y,x-1,y-1);
				g.drawLine(x+width-1,y,x+width,y-1);

				if (drawTitle!=null)
				{
					g.fillRect(x,y,width,12);
					boldFont.drawString(g,drawTitle,x+(width-boldFont.stringWidth(drawTitle)+1)/2,y,true);
				}
				if (help!=null)
				{
	//				symbolFont.drawChar(g,(char)255,x+width-13,y);
					int sx=x+width-12;
					g.setColor(bgColor);//PilrcEdit.bgColor));
					g.drawLine(sx,y+2,sx,y+7);
					g.drawLine(sx+1,y+1,sx+1,y+8);
					g.drawLine(sx+2,y,sx+2,y+9);
					g.drawLine(sx+7,y,sx+7,y+9);
					g.drawLine(sx+8,y+1,sx+8,y+8);
					g.drawLine(sx+9,y+2,sx+9,y+7);
          g.setColor(color?blue:Color.black);
					symbolFont.drawChar(g,(char)10,sx+3,y);
				}
			}
		}
		else
		if (drawTitle!=null&&r.y<y+15)
		{
			int swidth=boldFont.stringWidth(drawTitle)+4;
			g.fillRect(x+1,y,swidth-1,13);
			g.drawLine(x,y+1,x,y+12);
			g.drawLine(x+swidth,y+1,x+swidth,y+12);
			g.fillRect(x,y+13,width,2);
			boldFont.drawString(g,drawTitle,x+3,y+2,true);
		}
    g.setColor(Color.black);
		// hack to draw popped up lists on top of everything else
		if (poppedlist!=null)
		{
			g.translate(x,y);
      g.setColor(bgColor);
			poppedlist.clear(g);
      g.setColor(poppedlist.usable?Color.black:Color.lightGray);
			poppedlist.paint(g);
			g.translate(-x,-y);
		}
	}

  public int getID(String s)
  {
		return project.getConstant(s);
	}

  public Project getProject()
  {
		return project;
	}

  public FormObject getObject(int id)
  {
    //System.out.println("*********");
		FormObject obj=firstObject;
		while (obj!=null)
		{
      //System.out.println(id+" compare "+obj.getIDValue());
		  if (id==obj.getIDValue())
		  {
		    return obj;
			}
			obj=obj.next();
		}
		return null;
	}

	public boolean popupMenuItemSelected(String s)
	{
		if (s.equals("Make Copy"))
		{
			// do nothing
		}
		else
		if (s.equals("Delete"))
			project.deleteForm(this);
	  else
		if (s.equals("Auto Size"))
		{
			if (form.frame)
				setDimensions("2","2","156","156");
			else
				setDimensions("0","0","160","160");
		}
		else
  		return super.popupMenuItemSelected(s);
    return true;
	}

  public static String PROP_TITLE;
  public static String PROP_ID;
  public static String PROP_LEFT;
  public static String PROP_TOP;
  public static String PROP_WIDTH;
  public static String PROP_HEIGHT;
  public static String PROP_FRAME;
  public static String PROP_MODAL;
  public static String PROP_IFDEF;
  public static String PROP_IFNDEF;

  public void getProperties(Properties prop)
  {
		props=prop;
		prop.addProperty(PROP_TITLE,title,true);
		prop.addProperty(PROP_ID,id.getExpression());
		prop.addProperty(PROP_LEFT,xExp.getExpression());
		prop.addProperty(PROP_TOP,yExp.getExpression());
		prop.addProperty(PROP_WIDTH,widthExp.getExpression());
		prop.addProperty(PROP_HEIGHT,heightExp.getExpression());
		prop.addProperty(PROP_FRAME,boolStrings,frame?0:1);
		prop.addProperty(PROP_MODAL,boolStrings,modal?0:1);
    prop.addProperty(PROP_IFDEF,ifdef);
    prop.addProperty(PROP_IFNDEF,ifndef);
	}

  public void setTitle(String s)
  {
    if (s.equals(""))
      s=null;
    title=s;
    if (project!=null)
      project.titleChanged(this);
    setDirty(-2,0,width+4,15);
  }

  public boolean setProperty(String name,String value,int index)
  {
		if (name==PROP_TITLE)
		{
			setTitle(value);
		  return true;
		}
		else
		if (name==PROP_ID)
		{
			if (index==-2)
			{
				getProject().addID(value);
			}
			id.setExpression(value);
			project.titleChanged(this);
		}
	  else
		if (name==PROP_FRAME)
		{
			frame=index==0;
			if (!frame)
			{
				x-=2;
				y-=2;
				width+=4;
				height+=4;
			}
			else
			{
				x+=2;
				y+=2;
				width-=4;
				height-=4;
			}

      xExp.setValue(x);
      yExp.setValue(y);
      widthExp.setValue(width);
      heightExp.setValue(height);

			updatePositionProperties();

			if (!frame&&modal)
			{
				modal=false;
				props.setProperty(PROP_MODAL,1);
			}
		}
	  else
	  if (name==PROP_MODAL)
	  {
	    modal=index==0;
	    if (modal&&!frame)
	    {
	      props.setProperty(PROP_FRAME,0);
	      setProperty(PROP_FRAME,null,0);
			}
		}
		else
    if (name==PROP_IFDEF)
    {
      if (value.equals(""))
        value=null;
      ifdef=value;
      if (value!=null&&ifndef!=null)
      {
        ifndef=null;
        props.setProperty(PROP_IFNDEF,null);
      }
    }
    else
    if (name==PROP_IFNDEF)
    {
      if (value.equals(""))
        value=null;
      ifndef=value;
      if (value!=null&&ifdef!=null)
      {
        ifdef=null;
        props.setProperty(PROP_IFDEF,null);
      }
    }
    else
		{
			setAllDirty();
		  return super.setProperty(name,value,index);
		}
		evaluatePosition();
		setAllDirty();
		return true;
	}

  protected boolean parseOption(StreamTokenizer st) throws IOException
  {
		if (super.parseOption(st))
		  return true;

    int key=Keywords.get(this,st.sval);

    switch(key)
    {
      case FRAME: frame=true; break;
      case NOFRAME:
        if (!modal)
          frame=false;
        break;
      case MODAL: modal=frame=true; break;
      case SAVEBEHIND: saveBehind=true; break;
      case NOSAVEBEHIND: saveBehind=false; break;
      case HELPID: help=readNormalString(st); break;
      case MENUID: menu=readNormalString(st); break;
      case DEFAULTBTNID: defaultButton=readNormalString(st); break;
      case BEGIN:
        FormObject prev=null;
        st.nextToken();
        while (!(st.ttype==st.TT_WORD&&Keywords.get(this,st.sval)==END))
        {
          if (st.ttype==st.TT_WORD)
          {
            FormObject o=parseControl(st,prev,true);
            if (o!=null)
              prev=o;
          }
          st.nextToken();
        }
        clearTriggers();
        break;
      default: return false;
    }
		return true;
	}

  public FormObject parseControl(StreamTokenizer st,FormObject prev,boolean saveUnsupported) throws IOException
  {
    int key=Keywords.get(this,st.sval);
    switch(key)
    {
      case TITLE:
        setTitle(readQuotedString(st));
        return null;

      case POPUPLIST:
        String idTriggerSt=parseID(st);
        String idListSt=parseID(st);
        if (triggers==null)
          triggers=new Vector();
        triggers.addElement(idTriggerSt);
        triggers.addElement(idListSt);
        //System.out.println("popuplist:"+idTriggerSt+":"+idListSt);
        return null;

      default:
        st.pushBack();
        FormObject o=FormObject.parseObject(this,prev,st);
        if (o!=null)
          add(prev=o);
        else
        if (saveUnsupported)
        {
          if (unsupported==null)
            unsupported=new StringBuffer();
          st.eolIsSignificant(true);
          st.lowerCaseMode(false);
          unsupported.append("  ").append(st.sval.toUpperCase());
          while(st.nextToken()!=st.TT_EOL)
            unsupported.append(' ').append(st.sval);
          unsupported.append(Preferences.newLine);
          st.lowerCaseMode(true);
          st.eolIsSignificant(false);
        }
        return o;
    }
  }

  public void clearTriggers()
  {
    if (triggers!=null)
    {
      for(int i=0,size=triggers.size();i<size;i+=2)
      {
        String idTriggerSt=(String)triggers.elementAt(i);
        String idListSt=(String)triggers.elementAt(i+1);
        int idtrigger=Expression.evaluate(idTriggerSt);
        int idlist=Expression.evaluate(idListSt);
        try
        {
          PopupTrigger pt=(PopupTrigger)getObject(idtrigger);
          List list=(List)getObject(idlist);
          if (pt!=null&&list!=null)
          {
            pt.setList(list);
          }
        }
        catch(ClassCastException e)
        {
          System.out.println("popuplist:"+e);
        }
      }
      triggers=null;
    }
  }

	public void write(PrintStream ps)
	{
    Keywords.setDefault(this);
    boolean endif=false;
    if (ifdef!=null&&!ifdef.equals(""))
    {
      ps.println(Keywords.get(IFDEF)+" "+ifdef);
      endif=true;
    }
    else
    if (ifndef!=null&&!ifndef.equals(""))
    {
      ps.println(Keywords.get(IFNDEF)+" "+ifndef);
      endif=true;
    }
		ps.print(Keywords.get(NAME));
    ps.print(id.toOutputString());
		ps.println(" "+Keywords.get(AT)+" ("+xExp+" "+yExp+" "+widthExp+" "+heightExp+")");
		if (modal)
		  ps.println(Keywords.get(MODAL));
		if (!frame)
		  ps.println(Keywords.get(NOFRAME));
		if (!saveBehind)
		  ps.println(Keywords.get(NOSAVEBEHIND));
		if (help!=null)
		  ps.println(Keywords.get(HELPID)+" "+help);
		if (menu!=null)
		  ps.println(Keywords.get(MENUID)+" "+menu);
		if (defaultButton!=null)
		  ps.println(Keywords.get(DEFAULTBTNID)+" "+defaultButton);
		ps.println(Keywords.get(BEGIN));
		if (title!=null)
		  ps.println("  "+Keywords.get(TITLE)+" "+TextTools.quote(title));
		FormObject obj=firstObject;
		while(obj!=null)
		{
		  ps.print("  ");
		  obj.write(ps);
		  obj=obj.next();
	  }
	  if (unsupported!=null)
  		ps.print(unsupported.toString());
		ps.println(Keywords.get(this,END));
    if (endif)
      ps.println(Keywords.get(this,ENDIF));
    ps.println();
	}

	public int autoWidth()
	{
		return 156;
	}

	public int autoHeight()
	{
		return 156;
	}

  public void addFormListener(FormListener listener)
  {
    if (listeners==null)
      listeners=new Vector();
    listeners.addElement(listener);
  }

  public void removeFormListener(FormListener listener)
  {
    if (listeners==null)
      listeners=new Vector();
    listeners.removeElement(listener);
  }

  private void notifyListeners()
  {
    if (listeners==null)
      return;
    for(Enumeration e=listeners.elements();e.hasMoreElements();)
    {
      ((FormListener)e.nextElement()).formControlStructureChanged(this);
    }
  }
}