/**
 * PilrcEdit - a graphical editor for PilRC files.
 * Copyright (C) 2000 Rob Nielsen
 * rnielsen@wn.com.au
 * http://www.wn.com.au/rnielsen/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 */
package pilrcedit;

import pilrcedit.control.*;
import java.io.*;
import java.awt.Graphics;
import java.awt.Cursor;
import java.awt.Rectangle;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

public abstract class FormObject extends Resource implements Cloneable
{
  public static final int NAME=1000;
  public static final int DISPLAY_NAME=1001;

  public static final int ENABLED=200;
  public static final int DISABLED=201;
  public static final int USABLE=202;
  public static final int NONUSABLE=203;
  public static final int LEFTANCHOR=204;
  public static final int RIGHTANCHOR=205;
  public static final int FONT=206;
  public static final int AT=207;

	/**
	 * The number of pixels in an equation before breaking the relation.
	 * ie if breakAmount is 10 and the equation is PREVBOTTOM+11, the equation
	 * will be replaced with an absolute pixel value
	 */
	int breakAmount=10;

  /**
   * The number of pixels needed to be away from a potential relative relationship
   * before it is automatically made.
   */
  protected static int joinAmount=5;

	public static String[] boolStrings;

  private static Hashtable objectTypes;
  private static String[] objectStrings;

  private static void createObjectTypes()
  {
		objectTypes=new Hashtable();
		objectStrings=new String[14];
		int i=0;
    addObjectType(new Button(),i++);
    addObjectType(new Label(),i++);
    addObjectType(new Pushbutton(),i++);
    addObjectType(new RepeatButton(),i++);
    addObjectType(new Checkbox(),i++);
    addObjectType(new PopupTrigger(),i++);
    addObjectType(new List(),i++);
    addObjectType(new SelectorTrigger(),i++);
    addObjectType(new Field(),i++);
    addObjectType(new FormBitmap(),i++);
    addObjectType(new Gadget(),i++);
    addObjectType(new GraffitiShift(),i++);
    addObjectType(new Table(),i++);
    addObjectType(new ScrollBar(),i++);
	}

  private static void addObjectType(FormObject o,int i)
  {
    objectStrings[i]=o.getDisplayName();
    objectTypes.put(o.getObjectType().toLowerCase(),o);
  }

  public static String[] getObjectTypes()
  {
    if (objectStrings==null)
      createObjectTypes();
		return objectStrings;
	}

  /** The label on this formobject (if applicable) eg. 'Ok' on a button */
  protected String label=getObjectType();

  /** The left coordinate of this control from the left of it's parent form */
  protected int x;

  /** The top coordinate of this control from the top of it's parent form */
  protected int y;

  /** The width of the control */
  protected int width;

  /** The height of the control */
  protected int height;

  /** The font object used by this control */
  Font font;

  // general flags for formobject properties
  protected boolean usable=true;
  protected boolean disabled;
  protected boolean leftAnchor;
  protected boolean rightAnchor;

  /** The expression for the x coordinate of this control */
  protected LocationExpression xExp=new LocationExpression();

  /** The expression for the y coordinate of this control */
  protected LocationExpression yExp=new LocationExpression();

  /** The expression for the width of this control */
  protected DimensionExpression widthExp=new DimensionExpression();

  /** The expression for the height of this control */
  protected DimensionExpression heightExp=new DimensionExpression();

  /**
   * Is our position dependent on our previous control?
   */
  protected boolean isDependent=false;

  /** The parent Form for this control */
  protected Form form=null;

  /** The preceding object in the current form */
  private FormObject prev=null;

  /** The next object in the current form */
  private FormObject next=null;

  /** The properties window if we are currently selected, null otherwise */
  protected Properties props;

  /** The bounds of the pixels that we affect on the screen (usually larger than (x,y,width,height)) */
  protected Rectangle bounds;

  /** Is this control visible on the form? */
  private boolean visible=true;

  /**
   * Creates a new FormObject of the given type
   * @param name the object type to create
   * @return a new formobject, or null if the name is invalid
   */
  public static FormObject createObjectByName(String name)
  {
		try
		{
      if (objectTypes==null)
        createObjectTypes();
			return (FormObject)objectTypes.get(name.toLowerCase()).getClass().newInstance();
		}
		catch(Exception e)
		{
			return null;
		}
	}

  public static String convertDisplayNameToObjectType(String name)
  {
    try
    {
      if (objectTypes==null)
        createObjectTypes();
      for(Enumeration e=objectTypes.elements();e.hasMoreElements();)
      {
        FormObject o=(FormObject)e.nextElement();
        if (o.getDisplayName()==name)
          return o.getObjectType();
      }
    }
    catch(Exception e)
    {
    }
    return null;
  }

  public static FormObject parseObject(Form form,FormObject prev,StreamTokenizer st) throws IOException
  {
		st.nextToken();
		FormObject o=createObjectByName(st.sval);
		if (o==null)
		  return null;
		o.form=form;
		o.setPrev(prev);
		o.parse(st);
		return o;
	}

  protected Object clone() throws CloneNotSupportedException
  {
		FormObject obj=(FormObject)super.clone();
		obj.id=new IDExpression("AUTOID");
		obj.xExp=new LocationExpression(xExp);
		obj.yExp=new LocationExpression(yExp);
		obj.widthExp=new DimensionExpression(widthExp);
		obj.heightExp=new DimensionExpression(heightExp);
		return obj;
	}

  /**
   * Makes a full clone of this object.
   */
  public FormObject copy(String objType)
  {
		try
		{
			FormObject obj;
			// if it's the same type, make a full clone
			if (objType==null||objType.equals(getObjectType()))
			{
				obj=(FormObject)clone();
			}
			else  // otherwise, only copy the basic stuff like label and position
			{
        obj=createObjectByName(objType);
				if (obj==null)
					return null;
				if (obj.hasLabel())
					obj.label=label;
		  }
		  obj.setPrev(null);
		  obj.setNext(null);
		  obj.form=null;
		  obj.isDependent=false;
		  obj.widthExp.setExpression(widthExp.getExpression());
			obj.heightExp.setExpression(heightExp.getExpression());

      form.insert(obj,this);
  		if (isDependent)
		  {
				obj.xExp.setExpression(xExp.getExpression());
				obj.yExp.setExpression(yExp.getExpression());
		  }
		  else
		  {
				if (y+height*2<form.height)
				{
					obj.xExp.setExpression(Keywords.get(xExp,xExp.PREVLEFT));
					obj.yExp.setExpression(Keywords.get(yExp,yExp.PREVBOTTOM)+"+1");
				}
				else
				{
					obj.xExp.setExpression(Keywords.get(xExp,xExp.PREVRIGHT)+"+1");
					obj.yExp.setExpression(Keywords.get(yExp,yExp.PREVTOP));
				}
			}
		  obj.evaluatePosition();

		  return obj;
	  }
	  catch(Exception e){e.printStackTrace();}
    return null;
	}

  public void disconnect(boolean bothDirections)
  {
    xExp.setValue(x);
    yExp.setValue(y);

    if (!widthExp.isAuto())
      widthExp.setValue(width);
    if (!heightExp.isAuto())
      heightExp.setValue(height);

		if (hasDependent()&&bothDirections)
		  next.disconnect(false);
		evaluatePosition();
		updatePositionProperties();
	}

  public void setDimensions(String x,String y,String width,String height)
  {
		if (x!=null)
		{
		  xExp.setExpression(x);
		  if (props!=null)
		    props.setProperty("Left",x);
		}
		if (y!=null)
		{
		  yExp.setExpression(y);
		  if (props!=null)
		    props.setProperty("Top",y);
	  }
		if (width!=null)
		{
		  widthExp.setExpression(width);
		  if (props!=null)
		    props.setProperty("Width",width);
		}
		if (height!=null)
		{
		  heightExp.setExpression(height);
		  if (props!=null)
		    props.setProperty("Height",height);
		}
		evaluatePosition();
	}

  /**
   * Called when the formobject we are dependent on changes size or position.
   */
  public void evaluatePosition()
  {
		Rectangle oldBounds=getBounds();
		bounds=null;

		// recalc width
		widthExp.evaluate();
		if (widthExp.isAuto())
		  width=autoWidth();
		else
		  width=widthExp.getValue();

		// recalc height
		heightExp.evaluate();
		if (heightExp.isAuto())
		  height=autoHeight();
		else
		  height=heightExp.getValue();

		// recalc x
		xExp.evaluate();
		switch (xExp.getAlignment())
		{
			case LocationExpression.ALIGN_LEFT: x=xExp.getValue(); break;
      case LocationExpression.ALIGN_CENTER: x=(form.width-width)/2; break;
      case LocationExpression.ALIGN_CENTER_AT: x=xExp.getValue()-width/2; break;
      case LocationExpression.ALIGN_RIGHT: x=xExp.getValue()-width; break;
		}

		// recalc y
		yExp.evaluate();
		switch (yExp.getAlignment())
		{
      case LocationExpression.ALIGN_TOP: y=yExp.getValue(); break;
      case LocationExpression.ALIGN_CENTER: y=(form.height-height)/2; break;
      case LocationExpression.ALIGN_CENTER_AT: y=yExp.getValue()-height/2; break;
      case LocationExpression.ALIGN_BOTTOM: y=yExp.getValue()-height; break;
		}

		isDependent=xExp.isDependent()||yExp.isDependent();

    // notify our form that we need to redraw
    if (form!=null)
      form.setDirty(oldBounds);

		Rectangle newBounds=getBounds();

		// if our bounds haven't changed, don't bother telling our dependent.
		if (newBounds.equals(oldBounds))
		  return;

		// we've moved so tell our form to redraw the new position too
    if (form!=null)
      form.setDirty(newBounds);

		// notify our dependent that they may need to change
		if (hasDependent())
			next.evaluatePosition();
	}

  /**
   * Returns a string representation of this formobject
   */
	public String toString()
	{
    StringBuffer sb=new StringBuffer();
    sb.append(getObjectType());
    if (hasLabel())
      sb.append(" - \"").append(label).append('"');
    if (!id.isAutoID())
    {
      sb.append(" - ");
      if (id.isAbsolute())
        sb.append(id);
      else
		    sb.append(id).append(" (").append(id.getValue()).append(")");
    }
    return sb.toString();
	}

  /**
   * Does this object contain the given point?
   */
  public boolean contains(int px,int py)
  {
		return (px>=x-1&&px<=x+width&&py>=y-1&&py<=y+height);
	}

  public Rectangle getBounds()
  {
	  if (bounds==null)
	    bounds=new Rectangle(x-2,y-2,width+5,height+5);
	  return bounds;
	}

  /**
   * Gets the cursor that should be displayed if the mouse
   * is positioned at the given point
   * @return a Cursor constant indicating an appropriate cursor.
   */
  public int getCursor(int px,int py)
  {
		if (hasWidthAndHeight())
		{
			boolean left=Math.abs(px-x)<=1;
			boolean top=Math.abs(py-y)<=1;
			boolean right=Math.abs(px-(x+width-1))<=1;
			boolean bottom=Math.abs(py-(y+height-1))<=1;
			if (top&&left)
				return Cursor.NW_RESIZE_CURSOR;
			if (top&&right)
				return Cursor.NE_RESIZE_CURSOR;
			if (top)
				return Cursor.N_RESIZE_CURSOR;
			if (bottom&&left)
				return Cursor.SW_RESIZE_CURSOR;
			if (bottom&&right)
				return Cursor.SE_RESIZE_CURSOR;
			if (bottom)
				return Cursor.S_RESIZE_CURSOR;
			if (left)
				return Cursor.W_RESIZE_CURSOR;
			if (right)
				return Cursor.E_RESIZE_CURSOR;
		}
	  return Cursor.MOVE_CURSOR;
	}

  /**
   * Resize or reposition this object.  Called from Editor when dragging an
   * object around.
   * @param dx the change in x coordinate
   * @param dy the change in y coordinate
   * @param dw the change in width
   * @param dh the change in height
   * @return a rectangle containing all pixels affected
   */
	public void transform(int dx,int dy,int dw,int dh)
	{
		//System.out.println("transform("+dx+","+dy+","+dw+","+dh+")");
		if (dx==0&&dy==0&&dw==0&&dh==0)
		  return;

		boolean autolink=Preferences.getBooleanPref("Autolink",true);

    Keywords.setCase(Keywords.PREFERENCES);
		// create the current bounds, if necessary
		getBounds();

		x+=dx;
		y+=dy;
		width+=dw;
		height+=dh;
		if (dx!=0||(xExp.isRightAligned()&&dw!=0)) // Update X string
		{
			if (x==form.width/2-(width+1)/2)
			{
			  xExp.setExpression(Keywords.get(xExp,xExp.ALIGN_CENTER));
			}
			else
			if (xExp.isCenterAligned())
				xExp.setValue(x);
			else
			{
				int add=xExp.modifyExpression(dx+(xExp.isRightAligned()?dw:0));
				if (add>breakAmount)
					xExp.setValue(x);
			}
		}
		if (dy!=0||(yExp.isBottomAligned()&&dh!=0)) // Update Y string
		{
			if (y==form.height/2-(height+1)/2)
			{
			  yExp.setExpression(Keywords.get(yExp,yExp.ALIGN_CENTER));
			}
			else
			if (yExp.isCenterAligned())
				yExp.setValue(y);
			else
			{
				int add=yExp.modifyExpression(dy+(yExp.isBottomAligned()?dh:0));
				if (add>breakAmount)
					yExp.setValue(y);
			}
		}

    // look for dependencies
    if (autolink)
    {
			FormObject newPrev=isDependent?prev:null;
			if (newPrev==null)
				if ((newPrev=form.getPossibleDependent(this))!=null&&newPrev!=prev)
					form.insert(this,newPrev);
			if (newPrev!=null)
			{
				int xofs,yofs;
				int txofs=xofs=x-(prev.x+prev.width);
				int xtype=0;
				if (Math.abs(txofs=x-prev.x)<Math.abs(xofs))
				{ xofs=txofs; xtype=1; }
				if (Math.abs(txofs=x+width-(prev.x+prev.width))<Math.abs(xofs))
				{ xofs=txofs; xtype=2; }
				if (Math.abs(txofs=x+width-prev.x)<Math.abs(xofs))
				{ xofs=txofs; xtype=3; }

				if (Math.abs(xofs)<=joinAmount)
				{
					int tyofs=yofs=y-(prev.y+prev.height);
					int ytype=0;
					if (Math.abs(tyofs=y-prev.y)<Math.abs(yofs))
					{ yofs=tyofs; ytype=1; }
					if (Math.abs(tyofs=y+height-(prev.y+prev.height))<Math.abs(yofs))
					{ yofs=tyofs; ytype=2; }
					if (Math.abs(tyofs=y+height-prev.y)<Math.abs(yofs))
					{ yofs=tyofs; ytype=3; }

					if (Math.abs(yofs)<=joinAmount)
					{
						if (((xtype==0||xtype==3)&&(ytype==1||ytype==2))||
								((xtype==1||xtype==2)&&(ytype==0||ytype==3)))
						{
							String xexp=null;
							String yexp=null;
              Keywords.setDefault(xExp);
							switch(xtype)
							{
								case 0: xexp=Keywords.get(xExp.PREVRIGHT); break;
								case 1: xexp=Keywords.get(xExp.PREVLEFT); break;
								case 2: xexp=Keywords.get(xExp.ALIGN_RIGHT)+Keywords.get(xExp.PREVRIGHT); break;
                case 3: xexp=Keywords.get(xExp.ALIGN_RIGHT)+Keywords.get(xExp.PREVLEFT); break;
							}
							switch(ytype)
							{
                case 0: yexp=Keywords.get(yExp.PREVBOTTOM); break;
                case 1: yexp=Keywords.get(yExp.PREVTOP); break;
                case 2: yexp=Keywords.get(yExp.ALIGN_BOTTOM)+Keywords.get(yExp.PREVBOTTOM); break;
                case 3: yexp=Keywords.get(yExp.ALIGN_BOTTOM)+Keywords.get(yExp.PREVTOP); break;
							}
							if (!xExp.isCenterAligned())
							{
								xExp.setExpression(xexp);
								xExp.modifyExpression(xofs);
							}
							if (!yExp.isCenterAligned())
							{
								yExp.setExpression(yexp);
								yExp.modifyExpression(yofs);
							}
						}
					}
				}
			}
		}

		if (dw!=0) // Update width string
		{
			if (width==autoWidth())
			  widthExp.setAuto();
			else
			if (!widthExp.isAbsolute())
			{
				if ((isDependent&&Keywords.get(widthExp,widthExp.getExpression())==widthExp.PREVWIDTH&&
				   ((x!=prev.x&&y!=prev.y)||width!=prev.width))||
				   (widthExp.isAuto()))
				{
					widthExp.setValue(width);
				}
				else
					widthExp.modifyExpression(dw);
			}
			else
			if (autolink&&isDependent&&(x==prev.x||y==prev.y)&&width==prev.width&&
 				 (widthExp.isAbsolute()||
         Keywords.get(widthExp,prev.widthExp.getExpression())==widthExp.PREVWIDTH))
 				widthExp.setExpression(Keywords.get(widthExp.PREVWIDTH));
 		  else
 		    widthExp.setValue(width);
		}
		if (dh!=0) // Update height string
		{
			if (height==autoHeight())
			  heightExp.setAuto();
			else
			if (!heightExp.isAbsolute())
			{
				if ((isDependent&&Keywords.get(heightExp,heightExp.getExpression())==heightExp.PREVHEIGHT&&
				   ((x!=prev.x&&y!=prev.y)||height!=prev.height))||
				   (heightExp.isAuto()))
					heightExp.setValue(height);
				else
					heightExp.modifyExpression(dh);
			}
			else
			if (autolink&&isDependent&&(x==prev.x||y==prev.y)&&height==prev.height&&
			   (heightExp.isAbsolute()||
          Keywords.get(heightExp,prev.heightExp.getExpression())==heightExp.PREVHEIGHT))
			  heightExp.setExpression(Keywords.get(heightExp,heightExp.PREVHEIGHT));
			else
			  heightExp.setValue(height);
		}

		evaluatePosition();
    form.setDirty(getBounds());

    updatePositionProperties();
    Keywords.setCase(Keywords.MIXED_CASE);
	}

  ///////////////////////////////////////////////////////////
  // Popup Menu
  ///////////////////////////////////////////////////////////

  public void getPopupMenuItems(Vector v)
  {
		v.addElement("Make Copy");
		v.addElement("Delete");
		v.addElement("Auto Size");
		v.addElement("Disconnect");
	}

	public boolean popupMenuItemSelected(String s)
	{
		if (s.equals("Make Copy"))
		{
			FormObject obj=copy(null);
			if (obj!=null)
			  obj.requestFocus();
		}
		else
		if (s.equals("Delete"))
		{
			form.remove(this);
			form.requestFocus(next);
		}
		else
		if (s.equals("Auto Size"))
			setDimensions(null,null,"AUTO","AUTO");
		else
		if (s.equals("Disconnect"))
			disconnect(true);
		else
  		return false;
    return true;
	}

  ///////////////////////////////////////////////////////////
  // Properties
  ///////////////////////////////////////////////////////////

  // property names
  public static String PROP_LABEL;
  public static String PROP_ID;
  public static String PROP_LEFT;
  public static String PROP_TOP;
  public static String PROP_WIDTH;
  public static String PROP_HEIGHT;
  public static String PROP_FONT;
  public static String PROP_USABLE;
  public static String PROP_DISABLED;

  /**
   * Sets the properties of this object into the given Properties object
   * using the prop.addProperty() methods.
   */
  public void getProperties(Properties prop)
  {
		props=prop;
		if (hasLabel())
  		prop.addProperty(PROP_LABEL,label,true);
		prop.addProperty(PROP_ID,id.getExpression());
		prop.addProperty(PROP_LEFT,xExp.getExpression());
		prop.addProperty(PROP_TOP,yExp.getExpression());
		if (hasWidthAndHeight())
		{
			prop.addProperty(PROP_WIDTH,widthExp.getExpression());
			prop.addProperty(PROP_HEIGHT,heightExp.getExpression());
		}
		if (font!=null)
  		prop.addProperty(PROP_FONT,Font.getFontStrings(),font.getIndex());
		prop.addProperty(PROP_USABLE,boolStrings,usable?0:1);
		prop.addProperty(PROP_DISABLED,boolStrings,disabled?0:1);
	}

  /**
   * Called when another object is selected.  Changes should no longer be passed on
   * to the properties object.
   */
  public void propertiesClosed()
  {
		props=null;
	}

  public boolean setProperty(String name,String value,int index)
  {
		if (index<=-1)
		{
			if (name==PROP_LABEL)
			{
				label=value;
			}
			else
			if (name==PROP_ID)
      {
				if (index==-2)
				{
					getProject().addID(value);
				}
			  id.setExpression(value);
			}
			else
			if (name==PROP_LEFT)
				xExp.setExpression(value);
			else
			if (name==PROP_TOP)
				yExp.setExpression(value);
			else
			if (name==PROP_WIDTH)
				widthExp.setExpression(value);
			else
			if (name==PROP_HEIGHT)
				heightExp.setExpression(value);
		  else
		    return false;
	  }
		else
		if (name==PROP_FONT)
			font=Font.getFontAtIndex(index);
		else
		if (name==PROP_USABLE)
			usable=index==0;
	  else
		if (name==PROP_DISABLED)
			disabled=index==0;
		else
		  return false;
		evaluatePosition();
		return true;
	}

  protected void updatePositionProperties()
  {
		if (props!=null)
		{
			props.setProperty(PROP_LEFT,xExp.getExpression());
			props.setProperty(PROP_TOP,yExp.getExpression());
			props.setProperty(PROP_WIDTH,widthExp.getExpression());
			props.setProperty(PROP_HEIGHT,heightExp.getExpression());
		}
	}

  ///////////////////////////////////////////////////////////
  // Parsing
  ///////////////////////////////////////////////////////////

  public void parse(StreamTokenizer st) throws IOException
  {
    label=readQuotedString(st);
    readID(st);
    readLocation(st);
    parseOptions(st);
    evaluatePosition();
	}

  protected void parseOptions(StreamTokenizer st) throws IOException
  {
    usable=true;
    disabled=false;
    st.nextToken();
    while (st.sval!=null)
    {
			if (!parseOption(st))
			{
				st.pushBack();
			  return;
			}
			st.nextToken();
	  }
  }

  protected boolean parseOption(StreamTokenizer st) throws IOException
  {
		//System.out.println("st.sval="+st.sval);
    int key=Keywords.get(this,st.sval);
    switch(key)
    {
      case USABLE: usable=true; break;
      case NONUSABLE: usable=false; break;
      case DISABLED: disabled=true; break;
      case LEFTANCHOR: leftAnchor=true; break;
      case RIGHTANCHOR: rightAnchor=true; break;
      case FONT: setFont(Expression.evaluate(readNormalString(st))); break;
      default: return false;
    }
		return true;
  }

  protected boolean readLocation(StreamTokenizer st) throws IOException
  {
		st.ordinaryChar('(');
		st.nextToken();
    if (Keywords.get(this,st.sval)==AT)
      st.nextToken();
    st.lowerCaseMode(false);
    boolean bracketed=false;
    st.wordChars('(','(');
    if (st.ttype=='(')
    {
      st.nextToken();
      bracketed=true;
		}
    xExp.setExpression(st.sval);
    st.nextToken();
    if (bracketed&&!hasWidthAndHeight()&&st.sval.charAt(st.sval.length()-1)==')')
      yExp.setExpression(st.sval.substring(0,st.sval.length()-1));
    else
      yExp.setExpression(st.sval);
    if(hasWidthAndHeight())
    {
			st.nextToken();
			widthExp.setExpression(st.sval);
      st.nextToken();
      if (bracketed&&st.sval.charAt(st.sval.length()-1)==')')
        heightExp.setExpression(st.sval.substring(0,st.sval.length()-1));
      else
        heightExp.setExpression(st.sval);
    }
    else
    {
			widthExp.setAuto();
			heightExp.setAuto();
		}
    st.lowerCaseMode(true);
    st.ordinaryChar(')');
    st.nextToken();
    if (st.ttype!=')')
      st.pushBack();
    st.wordChars(')',')');
    //System.out.println("readLocation("+x+","+y+","+width+","+height+")");
    return true;
  }

  ///////////////////////////////////////////////////////////
  // Writing
  ///////////////////////////////////////////////////////////

  public void write(PrintStream ps)
  {
		StringBuffer sb=new StringBuffer();
		sb.append(Keywords.get(this,NAME));
		if (hasLabel())
			sb.append(' ').append(TextTools.quote(label));
    sb.append(id.toOutputString());
		sb.append(' ').append(Keywords.get(this,AT)).append(" (").append(xExp).append(' ').append(yExp);
		if (hasWidthAndHeight())
		{
			sb.append(' ').append(widthExp);
			sb.append(' ').append(heightExp);
		}
		sb.append(')');
		writeOptions(sb);
		ps.println(sb.toString());
	}

	protected void writeOptions(StringBuffer sb)
	{
    Keywords.setDefault(FormObject.class);
		if (font!=null)
		  sb.append(' ').append(Keywords.get(FONT)).append(' ').append(font.getID());
		if (!usable)
		  sb.append(' ').append(Keywords.get(NONUSABLE));
		if (disabled)
		  sb.append(' ').append(Keywords.get(DISABLED));
		if (leftAnchor)
		  sb.append(' ').append(Keywords.get(LEFTANCHOR));
		else
		if (rightAnchor)
		  sb.append(' ').append(Keywords.get(RIGHTANCHOR));
	}


  ///////////////////////////////////////////////////////////
  // Stupid little methods
  ///////////////////////////////////////////////////////////

  /**
   * Called when the component has been double clicked on.  Used for
   * Lists and PopupTriggers.
   */
  public void doubleClicked()
  {
	}

  public void requestFocus()
  {
		form.requestFocus(this);
	}

  public String getID()
  {
		return id.getExpression();
	}

  protected void setFont(int i)
  {
		font=Font.getFont(i);
	}

  public Font getFont()
  {
		if (font!=null)
		  return font;
		setFont(0);
		return font;
	}

  public int autoWidth()
  {
		return getFont().stringWidth(TextTools.unescape(label));
	}

	public int autoHeight()
	{
		return getFont().getHeight()+1;
	}

  public void clear(Graphics g)
  {
		g.fillRect(x,y,width,height);
	}

  public abstract void paint(Graphics g);

  public String getObjectType()
  {
    return Keywords.get(this,NAME);
  }

  public String getDisplayName()
  {
    return Keywords.get(this,DISPLAY_NAME);
  }

  public FormObject next()
  {
		return next;
	}

	public void setNext(FormObject obj)
	{
		next=obj;
		if (obj!=null)
		  obj.prev=this;
	}

	public FormObject prev()
	{
		return prev;
	}

	public void setPrev(FormObject obj)
	{
		prev=obj;
		if (obj!=null)
		  obj.next=this;
    else
      disconnect(false);
		xExp.setPrevious(obj);
		yExp.setPrevious(obj);
		widthExp.setPrevious(obj);
		heightExp.setPrevious(obj);
	}

  public boolean hasWidthAndHeight()
  {
		return true;
	}

  public boolean hasLabel()
  {
		return true;
	}

	public int getXAlignment()
	{
		return xExp.getAlignment();
	}

	public int getYAlignment()
	{
		return yExp.getAlignment();
	}

  public void setVisible(boolean b)
  {
		if (b!=visible)
		{
		  form.setDirty(getBounds());
		  visible=b;
		}
	}

  public boolean isVisible()
  {
		return visible;
	}

  public boolean hasDependent()
  {
		return next!=null&&next.isDependent;
	}

  public Project getProject()
  {
		return form.getProject();
	}

  public String getPositionString()
  {
		return "("+xExp+","+yExp+","+widthExp+","+heightExp+")";
	}

	public void setLabel(String s)
	{
		this.label=s;
	}
}