/**
 * PilrcEdit - a graphical editor for PilRC files.
 * Copyright (C) 2000 Rob Nielsen
 * rnielsen@wn.com.au
 * http://www.wn.com.au/rnielsen/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 */
package pilrcedit.control;

import pilrcedit.*;
import java.awt.Graphics;
import java.awt.Color;
import java.io.*;

public class Field extends FormObject
{
  public static final int EDITABLE=1;
  public static final int NONEDITABLE=2;
  public static final int UNDERLINED=3;
  public static final int SINGLELINE=4;
  public static final int MULTIPLELINES=5;
  public static final int DYNAMICSIZE=6;
  public static final int LEFTALIGN=7;
  public static final int HASSCROLLBAR=8;
  public static final int RIGHTALIGN=9;
  public static final int MAXCHARS=10;
  public static final int AUTOSHIFT=11;
  public static final int NUMERIC=12;

  public static String[] ALIGN_TYPES;
  public static String[] LINE_TYPES;
  public static String[] SIZE_TYPES;

  public static String PROP_ALIGN;
  public static String PROP_EDITABLE;
  public static String PROP_UNDERLINED;
  public static String PROP_LINES;
  public static String PROP_SIZE;
  public static String PROP_MAXCHARS;
  public static String PROP_AUTOSHIFT;
  public static String PROP_NUMERIC;
  public static String PROP_SCROLLBAR;

	boolean leftAlign=true;
	boolean editable=true;
	boolean underlined=true;
	boolean multiline=false;
	boolean dynamicSize=false;
	Expression maxChars=new Expression();
	boolean autoShift=false;
	boolean numeric=false;
	boolean scrollBar=false;

  protected Object clone() throws CloneNotSupportedException
  {
		Field f=(Field)super.clone();
		f.maxChars=new Expression(maxChars);
		return f;
	}

  public boolean hasLabel()
  {
		return false;
	}

	public void parse(StreamTokenizer st) throws IOException
	{
    underlined=false;
    readID(st);
    readLocation(st);
    parseOptions(st);
    evaluatePosition();
	}

  protected boolean parseOption(StreamTokenizer st) throws IOException
  {
		if (super.parseOption(st))
		  return true;

    int key=Keywords.get(this,st.sval);

    switch(key)
    {
      case LEFTALIGN: leftAlign=true; break;
      case RIGHTALIGN: leftAlign=false; break;
      case EDITABLE: editable=true; break;
      case NONEDITABLE: editable=false; break;
      case UNDERLINED: underlined=true; break;
      case SINGLELINE: multiline=false; break;
      case MULTIPLELINES: multiline=true; break;
      case DYNAMICSIZE: dynamicSize=true; break;
      case MAXCHARS: maxChars.setExpression(readNormalString(st)); break;
      case AUTOSHIFT: autoShift=true; break;
      case NUMERIC: numeric=true; break;
      case HASSCROLLBAR: scrollBar=true; break;
      default: return false;
    }
		return true;
  }

  public void getProperties(Properties prop)
  {
		super.getProperties(prop);
		prop.addProperty(PROP_ALIGN,ALIGN_TYPES,leftAlign?0:1);
		prop.addProperty(PROP_EDITABLE,boolStrings,editable?0:1);
		prop.addProperty(PROP_UNDERLINED,boolStrings,underlined?0:1);
		prop.addProperty(PROP_LINES,LINE_TYPES,multiline?1:0);
		prop.addProperty(PROP_SIZE,SIZE_TYPES,dynamicSize?0:1);
		prop.addProperty(PROP_MAXCHARS,maxChars.getExpression());
		prop.addProperty(PROP_AUTOSHIFT,boolStrings,autoShift?0:1);
		prop.addProperty(PROP_NUMERIC,boolStrings,numeric?0:1);
		prop.addProperty(PROP_SCROLLBAR,boolStrings,scrollBar?0:1);
	}

  public boolean setProperty(String name,String value,int index)
  {
		if (name==PROP_ALIGN)
			leftAlign=index==0;
		else
		if (name==PROP_EDITABLE)
		  editable=index==0;
		else
		if (name==PROP_UNDERLINED)
		  underlined=index==0;
		else
		if (name==PROP_LINES)
		  multiline=index==1;
		else
		if (name==PROP_SIZE)
		  dynamicSize=index==0;
		else
		if (name==PROP_MAXCHARS)
		  maxChars.setExpression(value);
		else
		if (name==PROP_AUTOSHIFT)
		  autoShift=index==0;
		else
		if (name==PROP_NUMERIC)
		  numeric=index==0;
		else
		if (name==PROP_SCROLLBAR)
		  scrollBar=index==0;
		else
		  return super.setProperty(name,value,index);
		evaluatePosition();
		return true;
	}

	protected void writeOptions(StringBuffer sb)
	{
		super.writeOptions(sb);
    Keywords.setDefault(this);
		if (!leftAlign)
		  sb.append(' ').append(Keywords.get(RIGHTALIGN));
		if (!editable)
		  sb.append(' ').append(Keywords.get(NONEDITABLE));
		if (underlined)
		  sb.append(' ').append(Keywords.get(UNDERLINED));
		if (multiline)
		  sb.append(' ').append(Keywords.get(MULTIPLELINES));
		if (dynamicSize)
		  sb.append(' ').append(Keywords.get(DYNAMICSIZE));
		if (maxChars.getValue()!=0)
		  sb.append(' ').append(Keywords.get(MAXCHARS)).append(' ').append(maxChars);
		if (autoShift)
		  sb.append(' ').append(Keywords.get(AUTOSHIFT));
		if (numeric)
		  sb.append(' ').append(Keywords.get(NUMERIC));
		if (scrollBar)
		  sb.append(' ').append(Keywords.get(HASSCROLLBAR));
	}

  public void paint(Graphics g)
  {
		if (underlined)
		{
      int fontHeight=getFont().getHeight();
			g.setColor(usable?Color.black:Color.lightGray);
			for(int j=1;j<=height/fontHeight;j++)
			{
				int y2=y+j*fontHeight-1;
				for(int i=x;i<x+width;i+=2)
					g.drawLine(i,y2,i,y2);
			}
			g.setColor(Color.black);
		}
  }
}